import logging
import time
from queue import Queue
from typing import TYPE_CHECKING, Optional

from transitions import State

from isar.models.communication.queues.queue_utils import check_for_event
from isar.services.utilities.threaded_request import ThreadedRequest
from robot_interface.models.exceptions.robot_exceptions import ErrorMessage

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


class Stop(State):
    def __init__(self, state_machine: "StateMachine") -> None:
        super().__init__(name="stop", on_enter=self.start, on_exit=self.stop)
        self.state_machine: "StateMachine" = state_machine
        self.logger = logging.getLogger("state_machine")
        self.events = self.state_machine.events
        self.stop_thread: Optional[ThreadedRequest] = None
        self._count_number_retries: int = 0
        self.signal_state_machine_to_stop = state_machine.signal_state_machine_to_stop

    def start(self) -> None:
        self.state_machine.update_state()
        self._run()

    def stop(self) -> None:
        if self.stop_thread:
            self.stop_thread.wait_for_thread()
        self.stop_thread = None
        self._count_number_retries = 0

    def _check_and_handle_failed_stop(self, event: Queue[ErrorMessage]) -> bool:
        error_message: Optional[ErrorMessage] = check_for_event(event)
        if error_message is not None:
            self.logger.warning(error_message.error_description)
            self.state_machine.mission_stopped()  # type: ignore
            return True
        return False

    def _check_and_handle_successful_stop(self, event: Queue[bool]) -> bool:
        if check_for_event(event):
            self.state_machine.mission_stopped()  # type: ignore
            return True
        return False

    def _run(self) -> None:
        while True:
            if self.signal_state_machine_to_stop.is_set():
                self.logger.info(
                    "Stopping state machine from %s state", self.__class__.__name__
                )
                break

            if self._check_and_handle_failed_stop(
                self.events.robot_service_events.mission_failed_to_stop
            ):
                break

            if self._check_and_handle_successful_stop(
                self.events.robot_service_events.mission_successfully_stopped
            ):
                break

            time.sleep(self.state_machine.sleep_time)



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine

from isar.apis.models.models import ControlMissionResponse
from robot_interface.models.mission.status import MissionStatus, TaskStatus


def pause_mission(state_machine: "StateMachine") -> bool:
    state_machine.logger.info("Pausing mission: %s", state_machine.current_mission.id)
    state_machine.current_mission.status = MissionStatus.Paused
    state_machine.current_task.status = TaskStatus.Paused

    paused_mission_response: ControlMissionResponse = (
        state_machine._make_control_mission_response()
    )
    state_machine.events.api_requests.pause_mission.output.put(paused_mission_response)

    state_machine.publish_mission_status()
    state_machine.publish_task_status(task=state_machine.current_task)

    state_machine.robot.pause()
    return True



import logging
import time
from typing import TYPE_CHECKING

from transitions import State

from isar.models.communication.queues.queue_utils import check_shared_state
from robot_interface.models.mission.status import RobotStatus

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


class Offline(State):
    def __init__(self, state_machine: "StateMachine") -> None:
        super().__init__(name="offline", on_enter=self.start, on_exit=self.stop)
        self.state_machine: "StateMachine" = state_machine
        self.logger = logging.getLogger("state_machine")
        self.shared_state = self.state_machine.shared_state
        self.signal_state_machine_to_stop = state_machine.signal_state_machine_to_stop

    def start(self) -> None:
        self.state_machine.update_state()
        self._run()

    def stop(self) -> None:
        return

    def _run(self) -> None:
        while True:
            if self.signal_state_machine_to_stop.is_set():
                self.logger.info(
                    "Stopping state machine from %s state", self.__class__.__name__
                )
                break

            robot_status: RobotStatus = check_shared_state(
                self.shared_state.robot_status
            )
            if robot_status == RobotStatus.BlockedProtectiveStop:
                transition = self.state_machine.robot_protective_stop_engaged  # type: ignore
                break
            elif robot_status != RobotStatus.Offline:
                transition = self.state_machine.robot_turned_online  # type: ignore
                break

            time.sleep(self.state_machine.sleep_time)

        transition()



import logging
import time
from typing import TYPE_CHECKING, Callable

from transitions import State

from isar.models.communication.queues.queue_utils import check_for_event

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


class Paused(State):
    def __init__(self, state_machine: "StateMachine") -> None:
        super().__init__(name="paused", on_enter=self.start)
        self.state_machine: "StateMachine" = state_machine
        self.logger = logging.getLogger("state_machine")
        self.events = self.state_machine.events
        self.signal_state_machine_to_stop = state_machine.signal_state_machine_to_stop

    def start(self) -> None:
        self.state_machine.update_state()
        self._run()

    def _run(self) -> None:
        transition: Callable
        while True:
            if self.signal_state_machine_to_stop.is_set():
                self.logger.info(
                    "Stopping state machine from %s state", self.__class__.__name__
                )
                break

            if check_for_event(self.events.api_requests.pause_mission.input):
                transition = self.state_machine.stop  # type: ignore
                break

            if check_for_event(self.events.api_requests.resume_mission.input):
                transition = self.state_machine.resume  # type: ignore
                break

            time.sleep(self.state_machine.sleep_time)

        transition()



import logging
import time
from copy import deepcopy
from queue import Queue
from threading import Event
from typing import TYPE_CHECKING, Optional, Tuple

from dependency_injector.wiring import inject
from transitions import State

from isar.config.settings import settings
from isar.models.communication.queues.queue_utils import check_for_event, trigger_event
from isar.services.utilities.threaded_request import ThreadedRequest
from robot_interface.models.exceptions.robot_exceptions import (
    ErrorMessage,
    RobotException,
    RobotRetrieveInspectionException,
)
from robot_interface.models.inspection.inspection import Inspection
from robot_interface.models.mission.mission import Mission
from robot_interface.models.mission.status import TaskStatus
from robot_interface.models.mission.task import InspectionTask, Task

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


class Monitor(State):
    @inject
    def __init__(self, state_machine: "StateMachine") -> None:
        super().__init__(name="monitor", on_enter=self.start, on_exit=self.stop)
        self.state_machine: "StateMachine" = state_machine

        self.logger = logging.getLogger("state_machine")
        self.events = self.state_machine.events

        self.awaiting_task_status: bool = False

        self.signal_state_machine_to_stop: Event = (
            state_machine.signal_state_machine_to_stop
        )

    def start(self) -> None:
        self.state_machine.update_state()
        self._run()

    def stop(self) -> None:
        self.state_machine.mission_ongoing = False
        return

    def _check_and_handle_stop_mission_event(self, event: Queue) -> bool:
        if check_for_event(event):
            self.state_machine.stop()  # type: ignore
            return True
        return False

    def _check_and_handle_pause_mission_event(self, event: Queue) -> bool:
        if check_for_event(event):
            self.state_machine.pause()  # type: ignore
            return True
        return False

    def _check_and_handle_mission_started_event(self, event: Queue) -> None:
        if not self.state_machine.mission_ongoing:
            if check_for_event(event):
                self.state_machine.mission_ongoing = True

    def _check_and_handle_mission_failed_event(self, event: Queue) -> bool:
        mission_failed: Optional[ErrorMessage] = check_for_event(event)
        if mission_failed is not None:
            self.state_machine.logger.warning(
                f"Failed to initiate mission "
                f"{str(self.state_machine.current_mission.id)[:8]} because: "
                f"{mission_failed.error_description}"
            )
            self.state_machine.current_mission.error_message = ErrorMessage(
                error_reason=mission_failed.error_reason,
                error_description=mission_failed.error_description,
            )

            self.state_machine.mission_failed_to_start()  # type: ignore
            return True
        return False

    def _check_and_handle_task_status_failed_event(self, event: Queue) -> bool:
        task_failure: Optional[ErrorMessage] = check_for_event(event)
        if task_failure is not None:
            self.awaiting_task_status = False
            self.state_machine.current_task.error_message = task_failure
            self.logger.error(
                f"Monitoring task {self.state_machine.current_task.id[:8]} failed "
                f"because: {task_failure.error_description}"
            )
            return self._handle_new_task_status(TaskStatus.Failed)
        elif not self.awaiting_task_status:
            trigger_event(
                self.events.state_machine_events.task_status_request,
                self.state_machine.current_task.id,
            )
            self.awaiting_task_status = True
        return False

    def _check_and_handle_task_status_event(self, event: Queue) -> bool:
        status: Optional[TaskStatus] = check_for_event(event)
        if status is not None:
            self.awaiting_task_status = False
            return self._handle_new_task_status(status)
        elif not self.awaiting_task_status:
            trigger_event(
                self.events.state_machine_events.task_status_request,
                self.state_machine.current_task.id,
            )
            self.awaiting_task_status = True
        return False

    def _handle_new_task_status(self, status: TaskStatus) -> bool:
        if self.state_machine.current_task is None:
            self.state_machine.iterate_current_task()

        self.state_machine.current_task.status = status

        if self._should_upload_inspections():
            get_inspection_thread = ThreadedRequest(self._queue_inspections_for_upload)
            get_inspection_thread.start_thread(
                deepcopy(self.state_machine.current_mission),
                deepcopy(self.state_machine.current_task),
                name="State Machine Get Inspections",
            )

        if self.state_machine.current_task.is_finished():
            self._report_task_status(self.state_machine.current_task)
            self.state_machine.publish_task_status(task=self.state_machine.current_task)

            self.state_machine.iterate_current_task()
            if self.state_machine.current_task is None:
                self.state_machine.mission_finished()  # type: ignore
                return True

            # Report and update next task
            self.state_machine.current_task.update_task_status()
            self.state_machine.publish_task_status(task=self.state_machine.current_task)
        return False

    def _run(self) -> None:
        self.awaiting_task_status = False
        while True:
            if self.signal_state_machine_to_stop.is_set():
                self.logger.info(
                    "Stopping state machine from %s state", self.__class__.__name__
                )
                break

            if self._check_and_handle_stop_mission_event(
                self.events.api_requests.stop_mission.input
            ):
                break

            if self._check_and_handle_pause_mission_event(
                self.events.api_requests.pause_mission.input
            ):
                break

            self._check_and_handle_mission_started_event(
                self.events.robot_service_events.mission_started
            )

            if self._check_and_handle_mission_failed_event(
                self.events.robot_service_events.mission_failed
            ):
                break

            if self._check_and_handle_task_status_failed_event(
                self.events.robot_service_events.task_status_failed
            ):
                break

            if self._check_and_handle_task_status_event(
                self.events.robot_service_events.task_status_updated
            ):
                break

            time.sleep(settings.FSM_SLEEP_TIME)

    def _queue_inspections_for_upload(
        self, mission: Mission, current_task: InspectionTask
    ) -> None:
        try:
            inspection: Inspection = self.state_machine.robot.get_inspection(
                task=current_task
            )
            if current_task.inspection_id != inspection.id:
                self.logger.warning(
                    f"The inspection_id of task ({current_task.inspection_id}) "
                    f"and result ({inspection.id}) is not matching. "
                    f"This may lead to confusions when accessing the inspection later"
                )

        except (RobotRetrieveInspectionException, RobotException) as e:
            self._set_error_message(e)
            self.logger.error(
                f"Failed to retrieve inspections because: {e.error_description}"
            )
            return

        except Exception as e:
            self.logger.error(
                f"Failed to retrieve inspections because of unexpected error: {e}"
            )
            return

        if not inspection:
            self.logger.warning(
                f"No inspection result data retrieved for task {str(current_task.id)[:8]}"
            )

        inspection.metadata.tag_id = current_task.tag_id

        message: Tuple[Inspection, Mission] = (
            inspection,
            mission,
        )
        self.state_machine.events.upload_queue.put(message)
        self.logger.info(
            f"Inspection result: {str(inspection.id)[:8]} queued for upload"
        )

    def _report_task_status(self, task: Task) -> None:
        if task.status == TaskStatus.Failed:
            self.logger.warning(
                f"Task: {str(task.id)[:8]} was reported as failed by the robot"
            )
        elif task.status == TaskStatus.Successful:
            self.logger.info(
                f"{type(task).__name__} task: {str(task.id)[:8]} completed"
            )

    def _should_upload_inspections(self) -> bool:
        if settings.UPLOAD_INSPECTIONS_ASYNC:
            return False

        return (
            self.state_machine.current_task.is_finished()
            and self.state_machine.current_task.status == TaskStatus.Successful
            and isinstance(self.state_machine.current_task, InspectionTask)
        )

    def _set_error_message(self, e: RobotException) -> None:
        error_message: ErrorMessage = ErrorMessage(
            error_reason=e.error_reason, error_description=e.error_description
        )
        self.state_machine.current_task.error_message = error_message



from collections import deque
from queue import Empty, Queue
from typing import TypeVar

T = TypeVar("T")


class StatusQueue(Queue[T]):
    def __init__(self) -> None:
        super().__init__()

    def check(self) -> T:
        if not self._qsize():
            raise Empty
        with self.mutex:
            queueList = list(self.queue)
            return queueList.pop()

    def update(self, item: T):
        with self.mutex:
            self.queue: deque[T] = deque()
            self.queue.append(item)



import logging
from copy import deepcopy
from http import HTTPStatus
from queue import Empty
from typing import Any, List

from dependency_injector.wiring import inject
from fastapi import HTTPException
from requests import HTTPError

from isar.apis.models.models import ControlMissionResponse
from isar.config.settings import settings
from isar.mission_planner.mission_planner_interface import (
    MissionNotFoundError,
    MissionPlannerError,
    MissionPlannerInterface,
)
from isar.models.communication.message import StartMissionMessage
from isar.models.communication.queues.events import APIRequests, Events, SharedState
from isar.models.communication.queues.queue_io import QueueIO
from isar.models.communication.queues.queue_timeout_error import QueueTimeoutError
from isar.services.utilities.queue_utilities import QueueUtilities
from isar.state_machine.states_enum import States
from robot_interface.models.mission.mission import Mission


class SchedulingUtilities:
    """
    Contains utility functions for scheduling missions from the API. The class handles
    required thread communication through queues to the state machine.
    """

    @inject
    def __init__(
        self,
        events: Events,
        shared_state: SharedState,
        mission_planner: MissionPlannerInterface,
        queue_timeout: int = settings.QUEUE_TIMEOUT,
    ):
        self.api_events: APIRequests = events.api_requests
        self.shared_state: SharedState = shared_state
        self.mission_planner: MissionPlannerInterface = mission_planner
        self.queue_timeout: int = queue_timeout
        self.logger = logging.getLogger("api")

    def get_state(self) -> States:
        """Return the current state of the state machine

        Raises
        ------
        HTTPException 500 Internal Server Error
            If the current state is not available on the queue
        """
        try:
            return self.shared_state.state.check()
        except Empty:
            error_message: str = (
                "Internal Server Error - Current state of the state machine is unknown"
            )
            self.logger.error(error_message)
            raise HTTPException(
                status_code=HTTPStatus.INTERNAL_SERVER_ERROR, detail=error_message
            )

    def get_mission(self, mission_id: str) -> Mission:
        """Get the mission with mission_id from the current mission planner

        Raises
        ------
        HTTPException 404 Not Found
            If requested mission with mission_id is not found
        HTTPException 500 Internal Server Error
            If for some reason the mission can not be returned
        """
        try:
            return self.mission_planner.get_mission(mission_id)
        except HTTPError as e:
            self.logger.error(e)
            raise HTTPException(status_code=e.response.status_code)
        except MissionNotFoundError as e:
            self.logger.error(e)
            raise HTTPException(
                status_code=HTTPStatus.NOT_FOUND,
                detail=f"Mission with id '{mission_id}' not found",
            )
        except MissionPlannerError as e:
            self.logger.error(e)
            raise HTTPException(
                status_code=HTTPStatus.INTERNAL_SERVER_ERROR,
                detail="Could not plan mission",
            )

    def verify_robot_capable_of_mission(
        self, mission: Mission, robot_capabilities: List[str]
    ) -> bool:
        """Verify that the robot is capable of performing this mission

        Raises
        ------
        HTTPException 400 Bad request
            If the robot is not capable of performing mission
        """
        missing_capabilities = {
            task.type for task in mission.tasks if task.type not in robot_capabilities
        }

        if missing_capabilities:
            error_message = (
                f"Bad Request - Robot is not capable of performing mission."
                f" Missing functionalities: {missing_capabilities}."
            )
            self.logger.error(error_message)
            raise HTTPException(
                status_code=HTTPStatus.BAD_REQUEST,
                detail=error_message,
            )

        return True

    def verify_state_machine_ready_to_receive_mission(self, state: States) -> bool:
        """Verify that the state machine is idle and ready to receive a mission

        Raises
        ------
        HTTPException 409 Conflict
            If state machine is not idle and therefore can not start a new mission
        """
        if state != States.Idle:
            error_message = f"Conflict - Robot is not idle - State: {state}"
            self.logger.warning(error_message)
            raise HTTPException(status_code=HTTPStatus.CONFLICT, detail=error_message)

        return True

    def start_mission(
        self,
        mission: Mission,
    ) -> None:
        """Start mission

        Raises
        ------
        HTTTPException 408 Request timeout
            If there is a timeout while communicating with the state machine
        """
        try:
            self._send_command(
                StartMissionMessage(mission=deepcopy(mission)),
                self.api_events.start_mission,
            )
        except QueueTimeoutError:
            error_message = "Internal Server Error - Failed to start mission in ISAR"
            self.logger.error(error_message)
            raise HTTPException(
                status_code=HTTPStatus.INTERNAL_SERVER_ERROR, detail=error_message
            )
        self.logger.info("OK - Mission started in ISAR")

    def pause_mission(self) -> ControlMissionResponse:
        """Pause mission

        Raises
        ------
        HTTTPException 408 Request timeout
            If there is a timeout while communicating with the state machine
        """
        try:
            return self._send_command(True, self.api_events.pause_mission)
        except QueueTimeoutError:
            error_message = "Internal Server Error - Failed to pause mission"
            self.logger.error(error_message)
            raise HTTPException(
                status_code=HTTPStatus.INTERNAL_SERVER_ERROR, detail=error_message
            )
        finally:
            self.logger.info("OK - Mission successfully paused")

    def resume_mission(self) -> ControlMissionResponse:
        """Resume mission

        Raises
        ------
        HTTTPException 408 Request timeout
            If there is a timeout while communicating with the state machine
        """
        try:
            return self._send_command(True, self.api_events.resume_mission)
        except QueueTimeoutError:
            error_message = "Internal Server Error - Failed to resume mission"
            self.logger.error(error_message)
            raise HTTPException(
                status_code=HTTPStatus.INTERNAL_SERVER_ERROR, detail=error_message
            )
        finally:
            self.logger.info("OK - Mission successfully resumed")

    def stop_mission(self) -> ControlMissionResponse:
        """Stop mission

        Raises
        ------
        HTTTPException 408 Request timeout
            If there is a timeout while communicating with the state machine
        """
        try:
            stop_mission_response: ControlMissionResponse = self._send_command(
                True, self.api_events.stop_mission
            )
        except QueueTimeoutError:
            error_message = "Internal Server Error - Failed to stop mission"
            self.logger.error(error_message)
            raise HTTPException(
                status_code=HTTPStatus.INTERNAL_SERVER_ERROR, detail=error_message
            )
        self.logger.info("OK - Mission successfully stopped")
        return stop_mission_response

    def _send_command(self, input: Any, queueio: QueueIO) -> Any:
        queueio.input.put(input)
        try:
            return QueueUtilities.check_queue(
                queueio.output,
                self.queue_timeout,
            )
        except QueueTimeoutError as e:
            QueueUtilities.clear_queue(queueio.input)
            self.logger.error("No output received for command to state machine")
            raise e



from typing import TYPE_CHECKING

from isar.models.communication.queues.queue_utils import trigger_event_without_data

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine

from isar.apis.models.models import ControlMissionResponse
from robot_interface.models.mission.status import MissionStatus, TaskStatus


def trigger_stop_mission_event(state_machine: "StateMachine") -> bool:
    trigger_event_without_data(state_machine.events.state_machine_events.stop_mission)
    return True


def stop_mission_cleanup(state_machine: "StateMachine") -> bool:
    if state_machine.current_mission is None:
        state_machine._queue_empty_response()
        state_machine.reset_state_machine()
        return True

    state_machine.current_mission.status = MissionStatus.Cancelled

    for task in state_machine.current_mission.tasks:
        if task.status in [
            TaskStatus.NotStarted,
            TaskStatus.InProgress,
            TaskStatus.Paused,
        ]:
            task.status = TaskStatus.Cancelled

    stopped_mission_response: ControlMissionResponse = (
        state_machine._make_control_mission_response()
    )
    state_machine.events.api_requests.stop_mission.output.put(stopped_mission_response)

    state_machine.publish_task_status(task=state_machine.current_task)
    state_machine._finalize()
    return True



from queue import Queue

from transitions import State

from isar.config.settings import settings
from isar.models.communication.queues.queue_io import QueueIO
from isar.models.communication.queues.status_queue import StatusQueue
from robot_interface.models.exceptions.robot_exceptions import ErrorMessage
from robot_interface.models.mission.mission import Mission
from robot_interface.models.mission.status import RobotStatus, TaskStatus
from robot_interface.models.mission.task import TASKS


class Events:
    def __init__(self) -> None:
        self.api_requests: APIRequests = APIRequests()
        self.state_machine_events: StateMachineEvents = StateMachineEvents()
        self.robot_service_events: RobotServiceEvents = RobotServiceEvents()

        self.upload_queue: Queue = Queue(maxsize=10)

        if settings.MQTT_ENABLED:
            self.mqtt_queue: Queue = Queue()


class APIRequests:
    def __init__(self) -> None:
        self.start_mission: QueueIO = QueueIO(input_size=1, output_size=1)
        self.stop_mission: QueueIO = QueueIO(input_size=1, output_size=1)
        self.pause_mission: QueueIO = QueueIO(input_size=1, output_size=1)
        self.resume_mission: QueueIO = QueueIO(input_size=1, output_size=1)


class StateMachineEvents:
    def __init__(self) -> None:
        self.start_mission: Queue[Mission] = Queue(maxsize=1)
        self.stop_mission: Queue[bool] = Queue(maxsize=1)
        self.pause_mission: Queue[bool] = Queue(maxsize=1)
        self.task_status_request: Queue[str] = Queue(maxsize=1)


class RobotServiceEvents:
    def __init__(self) -> None:
        self.task_status_updated: Queue[TaskStatus] = Queue(maxsize=1)
        self.task_status_failed: Queue[ErrorMessage] = Queue(maxsize=1)
        self.mission_started: Queue[bool] = Queue(maxsize=1)
        self.mission_failed: Queue[ErrorMessage] = Queue(maxsize=1)
        self.robot_status_changed: Queue[bool] = Queue(maxsize=1)
        self.mission_failed_to_stop: Queue[ErrorMessage] = Queue(maxsize=1)
        self.mission_successfully_stopped: Queue[bool] = Queue(maxsize=1)


class SharedState:
    def __init__(self) -> None:
        self.state: StatusQueue[State] = StatusQueue()
        self.robot_status: StatusQueue[RobotStatus] = StatusQueue()
        self.state_machine_current_task: StatusQueue[TASKS] = StatusQueue()



import logging
from http import HTTPStatus
from logging import Logger
from typing import List, Union

import click
import uvicorn
from dependency_injector.wiring import inject
from fastapi import FastAPI, Request, Security
from fastapi.middleware.cors import CORSMiddleware
from fastapi.routing import APIRouter
from opencensus.ext.azure.trace_exporter import AzureExporter
from opencensus.trace.attributes_helper import COMMON_ATTRIBUTES
from opencensus.trace.samplers import ProbabilitySampler
from opencensus.trace.span import SpanKind
from opencensus.trace.tracer import Tracer
from pydantic import AnyHttpUrl

from isar.apis.models.models import ControlMissionResponse, StartMissionResponse
from isar.apis.robot_control.robot_controller import RobotController
from isar.apis.schedule.scheduling_controller import SchedulingController
from isar.apis.security.authentication import Authenticator
from isar.config.configuration_error import ConfigurationError
from isar.config.keyvault.keyvault_error import KeyvaultError
from isar.config.keyvault.keyvault_service import Keyvault
from isar.config.settings import settings

HTTP_URL = COMMON_ATTRIBUTES["HTTP_URL"]
HTTP_STATUS_CODE = COMMON_ATTRIBUTES["HTTP_STATUS_CODE"]


class API:
    @inject
    def __init__(
        self,
        authenticator: Authenticator,
        scheduling_controller: SchedulingController,
        robot_controller: RobotController,
        keyvault: Keyvault,
        port: int = settings.API_PORT,
        azure_ai_logging_enabled: bool = settings.LOG_HANDLER_APPLICATION_INSIGHTS_ENABLED,
    ) -> None:
        self.authenticator: Authenticator = authenticator
        self.scheduling_controller: SchedulingController = scheduling_controller
        self.robot_controller: RobotController = robot_controller
        self.keyvault: Keyvault = keyvault
        self.host: str = "0.0.0.0"  # Locking uvicorn to use 0.0.0.0
        self.port: int = port
        self.azure_ai_logging_enabled: bool = azure_ai_logging_enabled

        self.logger: Logger = logging.getLogger("api")

        self.app: FastAPI = self._create_app()

    def get_app(self) -> FastAPI:
        return self.app

    def run_app(self) -> None:
        uvicorn.run(
            self.app,
            port=self.port,
            host=self.host,
            reload=False,
            log_config=None,
        )

    def _create_app(self) -> FastAPI:
        tags_metadata = [
            {
                "name": "Scheduler",
                "description": "Mission functionality",
            }
        ]
        app = FastAPI(
            openapi_tags=tags_metadata,
            on_startup=[self.authenticator.load_config, self._log_startup_message],
            swagger_ui_oauth2_redirect_url="/oauth2-redirect",
            swagger_ui_init_oauth={
                "usePkceWithAuthorizationCodeGrant": True,
                "clientId": settings.AZURE_CLIENT_ID,
            },
        )

        if self.authenticator.should_authenticate():
            backend_cors_origins: List[Union[str, AnyHttpUrl]] = [
                f"http://{self.host}:{self.port}"
            ]

            app.add_middleware(
                CORSMiddleware,
                allow_origins=[str(origin) for origin in backend_cors_origins],
                allow_credentials=True,
                allow_methods=["*"],
                allow_headers=["*"],
            )

        if self.azure_ai_logging_enabled:
            self._add_request_logging_middleware(app)

        app.include_router(router=self._create_scheduler_router())

        app.include_router(router=self._create_info_router())

        app.include_router(router=self._create_media_control_router())

        return app

    def _create_scheduler_router(self) -> APIRouter:
        router: APIRouter = APIRouter(tags=["Scheduler"])

        authentication_dependency: Security = Security(self.authenticator.get_scheme())

        router.add_api_route(
            path="/schedule/start-mission/{id}",
            endpoint=self.scheduling_controller.start_mission_by_id,
            methods=["POST"],
            deprecated=True,
            dependencies=[authentication_dependency],
            summary="Start a mission with id='id' from the current mission planner",
            responses={
                HTTPStatus.OK.value: {
                    "description": "Mission succesfully started",
                    "model": StartMissionResponse,
                },
                HTTPStatus.NOT_FOUND.value: {
                    "description": "Not found - Mission not found",
                },
                HTTPStatus.CONFLICT.value: {
                    "description": "Conflict - Invalid command in the current state",
                },
                HTTPStatus.INTERNAL_SERVER_ERROR.value: {
                    "description": "Internal Server Error - Current state of state machine unknown",
                },
            },
        )
        router.add_api_route(
            path="/schedule/start-mission",
            endpoint=self.scheduling_controller.start_mission,
            methods=["POST"],
            dependencies=[authentication_dependency],
            summary="Start the mission provided in JSON format",
            responses={
                HTTPStatus.OK.value: {
                    "description": "Mission succesfully started",
                    "model": StartMissionResponse,
                },
                HTTPStatus.UNPROCESSABLE_ENTITY.value: {
                    "description": "Invalid body - The JSON is incorrect",
                },
                HTTPStatus.CONFLICT.value: {
                    "description": "Conflict - Invalid command in the current state",
                },
                HTTPStatus.INTERNAL_SERVER_ERROR.value: {
                    "description": "Internal Server Error - Current state of state machine unknown",
                },
            },
        )
        router.add_api_route(
            path="/schedule/stop-mission",
            endpoint=self.scheduling_controller.stop_mission,
            methods=["POST"],
            dependencies=[authentication_dependency],
            summary="Stop the current mission",
            responses={
                HTTPStatus.OK.value: {
                    "description": "Mission succesfully stopped",
                    "model": ControlMissionResponse,
                },
                HTTPStatus.CONFLICT.value: {
                    "description": "Conflict - Invalid command in the current state",
                },
                HTTPStatus.INTERNAL_SERVER_ERROR.value: {
                    "description": "Internal Server Error - Current state of state machine unknown",
                },
            },
        )
        router.add_api_route(
            path="/schedule/pause-mission",
            endpoint=self.scheduling_controller.pause_mission,
            methods=["POST"],
            dependencies=[authentication_dependency],
            summary="Pause the current mission",
            responses={
                HTTPStatus.OK.value: {
                    "description": "Mission succesfully paused",
                    "model": ControlMissionResponse,
                },
                HTTPStatus.CONFLICT.value: {
                    "description": "Conflict - Invalid command in the current state",
                },
                HTTPStatus.INTERNAL_SERVER_ERROR.value: {
                    "description": "Internal Server Error - Current state of state machine unknown",
                },
            },
        )
        router.add_api_route(
            path="/schedule/resume-mission",
            endpoint=self.scheduling_controller.resume_mission,
            methods=["POST"],
            dependencies=[authentication_dependency],
            summary="Resume the currently paused mission - if any",
            responses={
                HTTPStatus.OK.value: {
                    "description": "Mission succesfully resumed",
                    "model": ControlMissionResponse,
                },
                HTTPStatus.CONFLICT.value: {
                    "description": "Conflict - Invalid command in the current state",
                },
                HTTPStatus.INTERNAL_SERVER_ERROR.value: {
                    "description": "Internal Server Error - Current state of state machine unknown",
                },
            },
        )
        router.add_api_route(
            path="/schedule/move_arm/{arm_pose_literal}",
            endpoint=self.scheduling_controller.start_move_arm_mission,
            methods=["POST"],
            dependencies=[authentication_dependency],
            summary="Move arm to the given arm pose literal",
            responses={
                HTTPStatus.OK.value: {
                    "description": "Move arm mission successfully started",
                },
                HTTPStatus.BAD_REQUEST.value: {
                    "description": "A move arm mission was scheduled on a robot that "
                    "does not support it or the input was incorrect",
                },
                HTTPStatus.CONFLICT.value: {
                    "description": "Conflict - Invalid command in the current state",
                },
                HTTPStatus.INTERNAL_SERVER_ERROR.value: {
                    "description": "Internal Server Error - Current state of state machine unknown",
                },
            },
        )

        return router

    def _create_info_router(self) -> APIRouter:
        router: APIRouter = APIRouter(tags=["Info"])

        authentication_dependency: Security = Security(self.authenticator.get_scheme())

        router.add_api_route(
            path="/info/robot-settings",
            endpoint=self.robot_controller.get_info,
            methods=["GET"],
            dependencies=[authentication_dependency],
            summary="Information about the robot-settings",
        )

        return router

    def _create_media_control_router(self) -> APIRouter:
        router: APIRouter = APIRouter(tags=["Media"])

        authentication_dependency: Security = Security(self.authenticator.get_scheme())

        router.add_api_route(
            path="/media/media-stream-config",
            endpoint=self.robot_controller.generate_media_config,
            methods=["GET"],
            dependencies=[authentication_dependency],
            summary="Generates a media stream connection config",
            responses={
                HTTPStatus.OK.value: {
                    "description": "Media stream was successfully generated",
                },
                HTTPStatus.NO_CONTENT.value: {
                    "description": "Robot has no media config",
                },
            },
        )

        return router

    def _log_startup_message(self) -> None:
        address_format = "%s://%s:%d/docs"
        message = f"Uvicorn running on {address_format} (Press CTRL+C to quit)"
        protocol = "http"
        color_message = (
            "Uvicorn running on "
            + click.style(address_format, bold=True)
            + " (Press CTRL+C to quit)"
        )
        self.logger.info(
            message,
            protocol,
            self.host,
            self.port,
            extra={"color_message": color_message},
        )

    def _add_request_logging_middleware(self, app: FastAPI) -> None:
        connection_string: str
        try:
            connection_string = self.keyvault.get_secret(
                "application-insights-connection-string"
            ).value
        except KeyvaultError:
            message: str = (
                "Missing connection string for Application Insights in key vault. "
            )
            self.logger.critical(message)
            raise ConfigurationError(message)

        @app.middleware("http")
        async def middlewareOpencensus(request: Request, call_next):
            tracer = Tracer(
                exporter=AzureExporter(connection_string=connection_string),
                sampler=ProbabilitySampler(1.0),
            )
            with tracer.span("main") as span:
                span.span_kind = SpanKind.SERVER

                response = await call_next(request)

                tracer.add_attribute_to_current_span(
                    attribute_key=HTTP_STATUS_CODE, attribute_value=response.status_code
                )
                tracer.add_attribute_to_current_span(
                    attribute_key=HTTP_URL, attribute_value=str(request.url)
                )

            return response



import logging
import time
from queue import Queue
from typing import TYPE_CHECKING, Optional

from transitions import State

from isar.models.communication.message import StartMissionMessage
from isar.models.communication.queues.queue_utils import (
    check_for_event,
    check_shared_state,
)
from isar.models.communication.queues.status_queue import StatusQueue
from robot_interface.models.mission.status import RobotStatus

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


class Idle(State):
    def __init__(self, state_machine: "StateMachine") -> None:
        super().__init__(name="idle", on_enter=self.start, on_exit=self.stop)
        self.state_machine: "StateMachine" = state_machine
        self.logger = logging.getLogger("state_machine")
        self.last_robot_status_poll_time: float = time.time()
        self.events = self.state_machine.events
        self.shared_state = self.state_machine.shared_state
        self.signal_state_machine_to_stop = state_machine.signal_state_machine_to_stop

    def start(self) -> None:
        self.state_machine.update_state()
        self._run()

    def stop(self) -> None:
        return

    def _check_and_handle_stop_mission_event(self, event: Queue) -> bool:
        if check_for_event(event):
            self.state_machine.stop()  # type: ignore
            return True
        return False

    def _check_and_handle_start_mission_event(
        self, event: Queue[StartMissionMessage]
    ) -> bool:
        start_mission: Optional[StartMissionMessage] = check_for_event(event)
        if start_mission:
            self.state_machine.start_mission(mission=start_mission.mission)
            self.state_machine.request_mission_start()  # type: ignore
            return True
        return False

    def _check_and_handle_robot_status_event(
        self, event: StatusQueue[RobotStatus]
    ) -> bool:
        robot_status: RobotStatus = check_shared_state(event)
        if robot_status == RobotStatus.Offline:
            self.state_machine.robot_turned_offline()  # type: ignore
            return True
        elif robot_status == RobotStatus.BlockedProtectiveStop:
            self.state_machine.robot_protective_stop_engaged()  # type: ignore
            return True
        return False

    def _run(self) -> None:
        while True:
            if self.signal_state_machine_to_stop.is_set():
                self.logger.info(
                    "Stopping state machine from %s state", self.__class__.__name__
                )
                break

            if self._check_and_handle_stop_mission_event(
                self.events.api_requests.stop_mission.input
            ):
                break

            if self._check_and_handle_start_mission_event(
                self.events.api_requests.start_mission.input
            ):
                break

            if self._check_and_handle_robot_status_event(
                self.shared_state.robot_status
            ):
                break

            time.sleep(self.state_machine.sleep_time)



import logging
from queue import Queue
from threading import Event
from typing import Optional

from dependency_injector.wiring import inject

from isar.models.communication.queues.events import (
    Events,
    RobotServiceEvents,
    SharedState,
    StateMachineEvents,
)
from isar.models.communication.queues.queue_utils import check_for_event
from isar.robot.robot_start_mission import RobotStartMissionThread
from isar.robot.robot_status import RobotStatusThread
from isar.robot.robot_stop_mission import RobotStopMissionThread
from isar.robot.robot_task_status import RobotTaskStatusThread
from robot_interface.robot_interface import RobotInterface


class Robot(object):
    @inject
    def __init__(
        self, events: Events, robot: RobotInterface, shared_state: SharedState
    ) -> None:
        self.logger = logging.getLogger("robot")
        self.state_machine_events: StateMachineEvents = events.state_machine_events
        self.robot_service_events: RobotServiceEvents = events.robot_service_events
        self.shared_state: SharedState = shared_state
        self.robot: RobotInterface = robot
        self.start_mission_thread: Optional[RobotStartMissionThread] = None
        self.robot_status_thread: Optional[RobotStatusThread] = None
        self.robot_task_status_thread: Optional[RobotTaskStatusThread] = None
        self.stop_mission_thread_thread: Optional[RobotStopMissionThread] = None
        self.signal_thread_quitting: Event = Event()

    def stop(self) -> None:
        self.signal_thread_quitting.set()
        if self.robot_status_thread is not None and self.robot_status_thread.is_alive():
            self.robot_status_thread.join()
        if (
            self.robot_task_status_thread is not None
            and self.robot_status_thread.is_alive()
        ):
            self.robot_task_status_thread.join()
        if (
            self.start_mission_thread is not None
            and self.robot_status_thread.is_alive()
        ):
            self.start_mission_thread.join()
        if (
            self.stop_mission_thread_thread is not None
            and self.stop_mission_thread_thread.is_alive()
        ):
            self.stop_mission_thread_thread.join()
        self.robot_status_thread = None
        self.robot_task_status_thread = None
        self.start_mission_thread = None

    def _check_and_handle_start_mission(self, event: Queue) -> None:
        start_mission = check_for_event(event)
        if start_mission is not None:
            if (
                self.start_mission_thread is not None
                and self.start_mission_thread.is_alive()
            ):
                self.logger.warning(
                    "Attempted to start mission while another mission was starting."
                )
                self.start_mission_thread.join()
            self.start_mission_thread = RobotStartMissionThread(
                self.robot_service_events,
                self.robot,
                self.signal_thread_quitting,
                start_mission,
            )
            self.start_mission_thread.start()

    def _check_and_handle_task_status_request(self, event: Queue[str]) -> None:
        task_id: str = check_for_event(event)
        if task_id:
            self.robot_task_status_thread = RobotTaskStatusThread(
                self.robot_service_events,
                self.robot,
                self.signal_thread_quitting,
                task_id,
            )
            self.robot_task_status_thread.start()

    def _check_and_handle_stop_mission(self, event: Queue) -> None:
        if check_for_event(event):
            if (
                self.stop_mission_thread_thread is not None
                and self.start_mission_thread.is_alive()
            ):
                return
            self.stop_mission_thread_thread = RobotStopMissionThread(
                self.robot_service_events, self.robot, self.signal_thread_quitting
            )
            self.stop_mission_thread_thread.start()

    def run(self) -> None:
        self.robot_status_thread = RobotStatusThread(
            self.robot, self.signal_thread_quitting, self.shared_state
        )
        self.robot_status_thread.start()

        while not self.signal_thread_quitting.wait(0):
            self._check_and_handle_start_mission(
                self.state_machine_events.start_mission
            )

            self._check_and_handle_task_status_request(
                self.state_machine_events.task_status_request
            )

            self._check_and_handle_stop_mission(self.state_machine_events.stop_mission)

        self.logger.info("Exiting robot service main thread")



import json
import logging
from collections import deque
from datetime import datetime, timezone
from threading import Event
from typing import Deque, List, Optional

from dependency_injector.wiring import inject
from transitions import Machine
from transitions.core import State

from isar.apis.models.models import ControlMissionResponse
from isar.config.settings import settings
from isar.mission_planner.task_selector_interface import (
    TaskSelectorInterface,
    TaskSelectorStop,
)
from isar.models.communication.queues.events import Events, SharedState
from isar.models.communication.queues.queue_utils import update_shared_state
from isar.state_machine.states.blocked_protective_stop import BlockedProtectiveStop
from isar.state_machine.states.idle import Idle
from isar.state_machine.states.monitor import Monitor
from isar.state_machine.states.off import Off
from isar.state_machine.states.offline import Offline
from isar.state_machine.states.paused import Paused
from isar.state_machine.states.stop import Stop
from isar.state_machine.states_enum import States
from isar.state_machine.transitions.fail_mission import (
    report_failed_mission_and_finalize,
)
from isar.state_machine.transitions.finish_mission import finish_mission
from isar.state_machine.transitions.pause import pause_mission
from isar.state_machine.transitions.resume import resume_mission
from isar.state_machine.transitions.start_mission import (
    initialize_robot,
    initiate_mission,
    put_start_mission_on_queue,
    set_mission_to_in_progress,
    trigger_start_mission_or_task_event,
)
from isar.state_machine.transitions.stop import (
    stop_mission_cleanup,
    trigger_stop_mission_event,
)
from isar.state_machine.transitions.utils import def_transition
from robot_interface.models.exceptions.robot_exceptions import ErrorMessage
from robot_interface.models.mission.mission import Mission
from robot_interface.models.mission.status import RobotStatus, TaskStatus
from robot_interface.models.mission.task import TASKS
from robot_interface.robot_interface import RobotInterface
from robot_interface.telemetry.mqtt_client import MqttClientInterface
from robot_interface.telemetry.payloads import (
    MissionPayload,
    RobotStatusPayload,
    TaskPayload,
)
from robot_interface.utilities.json_service import EnhancedJSONEncoder


class StateMachine(object):
    """Handles state transitions for supervisory robot control."""

    @inject
    def __init__(
        self,
        events: Events,
        shared_state: SharedState,
        robot: RobotInterface,
        mqtt_publisher: MqttClientInterface,
        task_selector: TaskSelectorInterface,
        sleep_time: float = settings.FSM_SLEEP_TIME,
        stop_robot_attempts_limit: int = settings.STOP_ROBOT_ATTEMPTS_LIMIT,
        transitions_log_length: int = settings.STATE_TRANSITIONS_LOG_LENGTH,
    ):
        """Initializes the state machine.

        Parameters
        ----------
        events : Events
            Events used for API and robot service communication.
        robot : RobotInterface
            Instance of robot interface.
        mqtt_publisher : MqttClientInterface
            Instance of MQTT client interface which has a publish function
        sleep_time : float
            Time to sleep in between state machine iterations.
        stop_robot_attempts_limit : int
            Maximum attempts to stop the robot when stop command is received
        transitions_log_length : int
            Length of state transition log list.

        """
        self.logger = logging.getLogger("state_machine")

        self.events: Events = events
        self.shared_state: SharedState = shared_state
        self.robot: RobotInterface = robot
        self.mqtt_publisher: Optional[MqttClientInterface] = mqtt_publisher
        self.task_selector: TaskSelectorInterface = task_selector

        self.signal_state_machine_to_stop: Event = Event()

        # List of states
        self.stop_state: State = Stop(self)
        self.paused_state: State = Paused(self)
        self.idle_state: State = Idle(self)
        self.monitor_state: State = Monitor(self)
        self.off_state: State = Off(self)
        self.offline_state: State = Offline(self)
        self.blocked_protective_stop: State = BlockedProtectiveStop(self)

        self.states: List[State] = [
            self.off_state,
            self.idle_state,
            self.monitor_state,
            self.stop_state,
            self.paused_state,
            self.offline_state,
            self.blocked_protective_stop,
        ]

        self.machine = Machine(self, states=self.states, initial="off", queued=True)
        self.machine.add_transitions(
            [
                {
                    "trigger": "start_machine",
                    "source": self.off_state,
                    "dest": self.idle_state,
                },
                {
                    "trigger": "pause",
                    "source": self.monitor_state,
                    "dest": self.paused_state,
                    "before": def_transition(self, pause_mission),
                },
                {
                    "trigger": "stop",
                    "source": [
                        self.idle_state,
                        self.monitor_state,
                        self.paused_state,
                    ],
                    "dest": self.stop_state,
                    "before": def_transition(self, trigger_stop_mission_event),
                },
                {
                    "trigger": "request_mission_start",
                    "source": self.idle_state,
                    "dest": self.monitor_state,
                    "prepare": def_transition(self, put_start_mission_on_queue),
                    "conditions": [
                        def_transition(self, initiate_mission),
                        def_transition(self, initialize_robot),
                    ],
                    "before": [
                        def_transition(self, set_mission_to_in_progress),
                        def_transition(self, trigger_start_mission_or_task_event),
                    ],
                },
                {
                    "trigger": "request_mission_start",
                    "source": self.idle_state,
                    "dest": self.idle_state,
                },
                {
                    "trigger": "mission_failed_to_start",
                    "source": self.monitor_state,
                    "dest": self.idle_state,
                    "before": def_transition(self, report_failed_mission_and_finalize),
                },
                {
                    "trigger": "resume",
                    "source": self.paused_state,
                    "dest": self.monitor_state,
                    "before": def_transition(self, resume_mission),
                },
                {
                    "trigger": "mission_finished",
                    "source": self.monitor_state,
                    "dest": self.idle_state,
                    "before": def_transition(self, finish_mission),
                },
                {
                    "trigger": "mission_stopped",
                    "source": self.stop_state,
                    "dest": self.idle_state,
                    "before": def_transition(self, stop_mission_cleanup),
                },
                {
                    "trigger": "robot_turned_offline",
                    "source": [self.idle_state, self.blocked_protective_stop],
                    "dest": self.offline_state,
                },
                {
                    "trigger": "robot_turned_online",
                    "source": self.offline_state,
                    "dest": self.idle_state,
                },
                {
                    "trigger": "robot_protective_stop_engaged",
                    "source": [self.idle_state, self.offline_state],
                    "dest": self.blocked_protective_stop,
                },
                {
                    "trigger": "robot_protective_stop_disengaged",
                    "source": self.blocked_protective_stop,
                    "dest": self.idle_state,
                },
            ]
        )

        self.stop_robot_attempts_limit: int = stop_robot_attempts_limit
        self.sleep_time: float = sleep_time

        self.current_mission: Optional[Mission] = None
        self.current_task: Optional[TASKS] = None

        self.mission_ongoing: bool = False

        self.current_state: State = States(self.state)  # type: ignore

        self.transitions_log_length: int = transitions_log_length
        self.transitions_list: Deque[States] = deque([], self.transitions_log_length)

    #################################################################################

    def _finalize(self) -> None:
        self.publish_mission_status()
        self.log_mission_overview(mission=self.current_mission)
        state_transitions: str = ", ".join(
            [
                f"\n  {transition}" if (i + 1) % 10 == 0 else f"{transition}"
                for i, transition in enumerate(list(self.transitions_list))
            ]
        )
        self.logger.info("State transitions:\n  %s", state_transitions)
        self.reset_state_machine()

    def begin(self):
        """Starts the state machine. Transitions into idle state."""
        self.to_idle()  # type: ignore

    def terminate(self):
        self.logger.info("Stopping state machine")
        self.signal_state_machine_to_stop.set()

    def iterate_current_task(self):
        if self.current_task.is_finished():
            try:
                self.current_task = self.task_selector.next_task()
                self.current_task.status = TaskStatus.InProgress
                self.publish_task_status(task=self.current_task)
            except TaskSelectorStop:
                # Indicates that all tasks are finished
                self.current_task = None
            self.send_task_status()

    def update_state(self):
        """Updates the current state of the state machine."""
        self.current_state = States(self.state)  # type: ignore
        update_shared_state(self.shared_state.state, self.current_state)
        self._log_state_transition(self.current_state)
        self.logger.info("State: %s", self.current_state)
        self.publish_status()

    def reset_state_machine(self) -> None:
        self.logger.info("Resetting state machine")
        self.current_task = None
        self.send_task_status()
        self.current_mission = None

    def start_mission(self, mission: Mission):
        """Starts a scheduled mission."""
        self.current_mission = mission

        self.task_selector.initialize(tasks=self.current_mission.tasks)

    def send_task_status(self):
        update_shared_state(
            self.shared_state.state_machine_current_task, self.current_task
        )

    def publish_mission_status(self) -> None:
        if not self.mqtt_publisher:
            return

        error_message: Optional[ErrorMessage] = None
        if self.current_mission:
            if self.current_mission.error_message:
                error_message = self.current_mission.error_message

        payload: MissionPayload = MissionPayload(
            isar_id=settings.ISAR_ID,
            robot_name=settings.ROBOT_NAME,
            mission_id=self.current_mission.id if self.current_mission else None,
            status=self.current_mission.status if self.current_mission else None,
            error_reason=error_message.error_reason if error_message else None,
            error_description=(
                error_message.error_description if error_message else None
            ),
            timestamp=datetime.now(timezone.utc),
        )

        self.mqtt_publisher.publish(
            topic=settings.TOPIC_ISAR_MISSION,
            payload=json.dumps(payload, cls=EnhancedJSONEncoder),
            qos=1,
            retain=True,
        )

    def publish_task_status(self, task: TASKS) -> None:
        """Publishes the task status to the MQTT Broker"""
        if not self.mqtt_publisher:
            return

        error_message: Optional[ErrorMessage] = None
        if task:
            if task.error_message:
                error_message = task.error_message

        payload: TaskPayload = TaskPayload(
            isar_id=settings.ISAR_ID,
            robot_name=settings.ROBOT_NAME,
            mission_id=self.current_mission.id if self.current_mission else None,
            task_id=task.id if task else None,
            status=task.status if task else None,
            task_type=task.type if task else None,
            error_reason=error_message.error_reason if error_message else None,
            error_description=(
                error_message.error_description if error_message else None
            ),
            timestamp=datetime.now(timezone.utc),
        )

        self.mqtt_publisher.publish(
            topic=settings.TOPIC_ISAR_TASK,
            payload=json.dumps(payload, cls=EnhancedJSONEncoder),
            qos=1,
            retain=True,
        )

    def publish_status(self) -> None:
        if not self.mqtt_publisher:
            return

        payload: RobotStatusPayload = RobotStatusPayload(
            isar_id=settings.ISAR_ID,
            robot_name=settings.ROBOT_NAME,
            status=self._current_status(),
            timestamp=datetime.now(timezone.utc),
        )

        self.mqtt_publisher.publish(
            topic=settings.TOPIC_ISAR_STATUS,
            payload=json.dumps(payload, cls=EnhancedJSONEncoder),
            qos=1,
            retain=True,
        )

    def _current_status(self) -> RobotStatus:
        if self.current_state == States.Idle:
            return RobotStatus.Available
        elif self.current_state == States.Offline:
            return RobotStatus.Offline
        elif self.current_state == States.BlockedProtectiveStop:
            return RobotStatus.BlockedProtectiveStop
        else:
            return RobotStatus.Busy

    def _log_state_transition(self, next_state) -> None:
        """Logs all state transitions that are not self-transitions."""
        self.transitions_list.append(next_state)

    def log_mission_overview(self, mission: Mission) -> None:
        """Log an overview of the tasks in a mission"""
        log_statements: List[str] = []
        for task in mission.tasks:
            log_statements.append(
                f"{type(task).__name__:<20} {str(task.id)[:8]:<32} -- {task.status}"
            )
        log_statement: str = "\n".join(log_statements)

        self.logger.info("Mission overview:\n%s", log_statement)

    def _make_control_mission_response(self) -> ControlMissionResponse:
        return ControlMissionResponse(
            mission_id=self.current_mission.id,
            mission_status=self.current_mission.status,
            task_id=self.current_task.id,
            task_status=self.current_task.status,
        )

    def _queue_empty_response(self) -> None:
        self.events.api_requests.stop_mission.output.put(
            ControlMissionResponse(
                mission_id="None",
                mission_status="None",
                task_id="None",
                task_status="None",
            )
        )


def main(state_machine: StateMachine):
    """Starts a state machine instance."""
    state_machine.begin()



from queue import Empty, Queue
from typing import Optional, TypeVar

from isar.models.communication.queues.status_queue import StatusQueue

T = TypeVar("T")


def trigger_event_without_data(queue: Queue[bool]) -> None:
    queue.put(True)


def trigger_event(queue: Queue[T], data: T) -> None:
    queue.put(data)


def check_shared_state(queue: StatusQueue[T]) -> Optional[T]:
    try:
        return queue.check()
    except Empty:
        return None


def update_shared_state(queue: StatusQueue[T], data: T) -> None:
    queue.update(data)


def check_for_event(queue: Queue[T]) -> Optional[T]:
    try:
        return queue.get(block=False)
    except Empty:
        return None
